"""
Configure Celery with the following features:
* Disable the results backend
* Ensure that tasks will never be lost, but tasks themselves must be idempotent
* Optimize the network connection to CloudAMQP

This requires the `celery` package to be installed.
"""

import celery.app.trace  # type: ignore[import-not-found]

from resonant_settings._env import env

# Assume AMQP.
CELERY_BROKER_URL: str = env.str("DJANGO_CELERY_BROKER_URL")

# Disable results backend, as this feature has too many weaknesses.
# The database should be used to communicate results of completed tasks.
CELERY_RESULT_BACKEND = None

# Only acknowledge a task being done after the function finishes.
# This provides safety against worker crashes, but adds the requirement
# that tasks must be idempotent (which is a best practice anyway).
# See: https://docs.celeryproject.org/en/stable/faq.html#should-i-use-retry-or-acks-late
CELERY_TASK_ACKS_LATE = True

# When a worker subprocess abruptly exists, assume it was is killed by the operating system for
# a cause which is intrinsic (e.g. a segfault or OOM) to the task it was running, so do not
# requeue. It's expected that the task wouldn't succeed if run again.
# This should not impact cases where the task fails due to extrinsic causes (e.g. the process
# supervisor sends a SIGKILL or the machine loses power), as we assume that the parent worker
# process will immediately die too (and not have a chance to requeue the task).
# See: https://docs.celeryproject.org/en/stable/userguide/tasks.html#tasks for more explanation
# of these tradeoffs.
# None of this affects warm shutdowns from a SIGTERM (which the process supervisor ought to
# send), as this just allows Celery to complete running tasks; see:
# https://docs.celeryproject.org/en/stable/userguide/workers.html#process-signals for reference.
# This is Celery's default.
CELERY_TASK_REJECT_ON_WORKER_LOST = False

# When a task fails due to an internally-raised exception or due to a timeout, do not requeue.
# It's expected that the task wouldn't succeed if run again.
# This is Celery's default.
CELERY_TASK_ACKS_ON_FAILURE_OR_TIMEOUT = True

# This is sensible behavior with TASKS_ACKS_LATE, this must be enabled to prevent warnings,
# and this will be Celery's default in 6.0.
CELERY_WORKER_CANCEL_LONG_RUNNING_TASKS_ON_CONNECTION_LOSS = True

# CloudAMQP-suggested settings
# https://www.cloudamqp.com/docs/celery.html
CELERY_BROKER_POOL_LIMIT = 1
CELERY_BROKER_HEARTBEAT = None
CELERY_BROKER_CONNECTION_TIMEOUT = 30
CELERY_EVENT_QUEUE_EXPIRES = 60

# Note, CELERY_WORKER settings could be different on each running worker.

# Do not prefetch, as the speed benefit for fast-running tasks may not be
# worth a potentially unfair allocation with slow-running tasks and
# multiple workers.
CELERY_WORKER_PREFETCH_MULTIPLIER = 1

# Accept the default of the number of CPU cores.
# Workers running memory-intensive tasks may need to decrease this.
CELERY_WORKER_CONCURRENCY: int | None = None

# Make the task received log message include the task's arguments.
celery.app.trace.LOG_RECEIVED = """\
Task %(name)s[%(id)s] received: (%(args)s, %(kwargs)s)\
"""

__all__ = [
    "CELERY_BROKER_URL",
    "CELERY_RESULT_BACKEND",
    "CELERY_TASK_ACKS_LATE",
    "CELERY_TASK_REJECT_ON_WORKER_LOST",
    "CELERY_TASK_ACKS_ON_FAILURE_OR_TIMEOUT",
    "CELERY_WORKER_CANCEL_LONG_RUNNING_TASKS_ON_CONNECTION_LOSS",
    "CELERY_BROKER_POOL_LIMIT",
    "CELERY_BROKER_HEARTBEAT",
    "CELERY_BROKER_CONNECTION_TIMEOUT",
    "CELERY_EVENT_QUEUE_EXPIRES",
    "CELERY_WORKER_PREFETCH_MULTIPLIER",
    "CELERY_WORKER_CONCURRENCY",
]



from typing import cast

from allauth.account import app_settings as allauth_settings
from django.contrib.auth import get_user_model
from django.contrib.auth.management.commands import createsuperuser as django_createsuperuser
from django.contrib.auth.models import AbstractUser
from django.core.management import BaseCommand
from django.db.models.signals import post_save

from resonant_settings.allauth_support import createsuperuser as allauth_support_createsuperuser
from resonant_settings.allauth_support.receiver import verify_email_address_on_user_post_save

"""
When Allauth is configured to use a User's `email` as the `username`, override the `createsuperuser`
management command to only prompt for an email address.
"""

username_required: bool | None = allauth_settings.SIGNUP_FIELDS.get("username", {}).get(
    "required", None
)

# If using email as username
if not username_required:
    # Expose the modified command
    Command: type[BaseCommand] = allauth_support_createsuperuser.Command
    user_model: type[AbstractUser] = allauth_support_createsuperuser.EmailAsUsernameProxyUser

else:
    # Expose the pristine upstream version of the command
    Command = django_createsuperuser.Command
    user_model = cast(type[AbstractUser], get_user_model())

# Always automatically verify email addresses of newly created superusers
post_save.connect(verify_email_address_on_user_post_save, sender=user_model)



"""
Configure django-allauth with the following features:
* Disable usernames for end users, using exclusively email addresses for login
* Require email verification
* Quality of life improvements for users

This requires the `django-allauth` package to be installed and requires
`resonant_settings.allauth_support` to be added to INSTALLED_APPS.
"""

from collections.abc import Sequence

# The sites framework requires this to be set.
# In the unlikely case where a database's pk sequence for the django_site table is not reset,
# the default site object could have a different pk. Then this will need to be overridden
# downstream.
SITE_ID = 1

AUTHENTICATION_BACKENDS: Sequence[str] = [
    # Django's built-in ModelBackend is not necessary, since all users will be
    # authenticated by their email address
    "allauth.account.auth_backends.AuthenticationBackend",
]

# see configuration documentation at
#   https://django-allauth.readthedocs.io/en/latest/configuration.html

# Require email verification
ACCOUNT_EMAIL_VERIFICATION = "mandatory"

# Use email as the identifier for login
ACCOUNT_LOGIN_METHODS = {"email"}
# Don't require a username, but make email required
ACCOUNT_SIGNUP_FIELDS = ["email*", "password1*", "password2*"]

# Set the username as the email
ACCOUNT_ADAPTER = "resonant_settings.allauth_support.adapter.EmailAsUsernameAccountAdapter"
ACCOUNT_USER_MODEL_USERNAME_FIELD = None

# Quality of life improvements, but may not work if the browser is closed
ACCOUNT_SESSION_REMEMBER = True
ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION = True
ACCOUNT_LOGIN_ON_PASSWORD_RESET = True

# Confirm URLs include a secret token, so CSRF safety isn't a concern
ACCOUNT_CONFIRM_EMAIL_ON_GET = True

# This will likely become the default in the future, but enable it now
ACCOUNT_PRESERVE_USERNAME_CASING = False

__all__ = [
    "SITE_ID",
    "AUTHENTICATION_BACKENDS",
    "ACCOUNT_EMAIL_VERIFICATION",
    "ACCOUNT_LOGIN_METHODS",
    "ACCOUNT_SIGNUP_FIELDS",
    "ACCOUNT_ADAPTER",
    "ACCOUNT_USER_MODEL_USERNAME_FIELD",
    "ACCOUNT_SESSION_REMEMBER",
    "ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION",
    "ACCOUNT_LOGIN_ON_PASSWORD_RESET",
    "ACCOUNT_CONFIRM_EMAIL_ON_GET",
    "ACCOUNT_PRESERVE_USERNAME_CASING",
]



import logging
from typing import Any

from allauth.account.models import EmailAddress
from django.contrib.auth.models import AbstractUser

logger = logging.getLogger(__file__)


def verify_email_address_on_user_post_save(
    sender: type[AbstractUser],
    instance: AbstractUser,
    created: bool,
    **kwargs: Any,
) -> None:
    """Automatically verify email addresses of newly created superusers."""
    # These should always be true, but it's a final sanity check
    if created and instance.is_superuser:
        # Django is less strict than Allauth about duplicate email addresses (Django lowercases
        # the domain portion or an address, Allauth lowercases the whole address). It's possible
        # an acceptable email address via createsuperuser would be refused as a duplicate via
        # Allauth.
        # It's also possible that in an mature database where users have multiple non-primary
        # emails, the new user's email matches an existing user's non-primary email, violating
        # uniqueness.
        # So, make a conservative effort at setting the email address as verified, since failure is
        # not critical and can be resolved by the user.
        email_address, created = EmailAddress.objects.get_or_create(
            email__iexact=instance.email,
            defaults={
                "user": instance,
                "email": instance.email,
                "verified": True,
                "primary": True,
            },
        )
        if not created:
            logger.warning(f'Could not automatically verify email address "{instance.email}".')



"""
Configure Django's email sending.

The following environment variables must be externally set:
* `DJANGO_EMAIL_URL`, as a URL for login to an STMP server, as parsed by `dj-email-url`. This
  typically will start with `submission:`. Special characters in passwords must be URL-encoded.
  See https://pypi.org/project/dj-email-url/ for full details.
* `DJANGO_DEFAULT_FROM_EMAIL`, as the default From address for outgoing email.
"""

from typing import Any

from resonant_settings._env import env

email_config: dict[str, Any] = env.email_url("DJANGO_EMAIL_URL")
globals().update(email_config)

DEFAULT_FROM_EMAIL: str = env.str("DJANGO_DEFAULT_FROM_EMAIL")
SERVER_EMAIL = DEFAULT_FROM_EMAIL


__all__ = [
    *email_config.keys(),
    "DEFAULT_FROM_EMAIL",
    "SERVER_EMAIL",
]



from collections.abc import Generator
from contextlib import contextmanager
from typing import Any


# From https://stackoverflow.com/a/38532086
@contextmanager
def temporarily_change_attributes(something: object, **kwargs: Any) -> Generator[None]:
    previous_values = {k: getattr(something, k) for k in kwargs}
    for k, v in kwargs.items():
        setattr(something, k, v)
    try:
        yield
    finally:
        for k, v in previous_values.items():
            setattr(something, k, v)



"""
Configure Django logging with the following features:
* Emit all logs to stdout
* Exclude favicons and static files from request and server logs
* Improve log formatting and add colorization

This requires the `rich` package to be installed.
"""

import logging

from django.http import HttpRequest


def _filter_favicon_requests(record: logging.LogRecord) -> bool:
    if record.name == "django.request":
        request: HttpRequest | None = getattr(record, "request", None)
        if request and request.path == "/favicon.ico":
            return False

    if (
        record.name == "django.server"
        and isinstance(record.args, tuple)
        and len(record.args) >= 1
        and str(record.args[0]).startswith("GET /favicon.ico ")
    ):
        return False

    return True


def _filter_static_requests(record: logging.LogRecord) -> bool:
    if (
        record.name == "django.server"
        and isinstance(record.args, tuple)
        and len(record.args) >= 1
        and str(record.args[0]).startswith("GET /static/")
    ):
        return False

    return True


LOGGING = {
    "version": 1,
    # Replace existing logging configuration
    "incremental": False,
    # This redefines all of Django's declared loggers, but most loggers are implicitly
    # declared on usage, and should not be disabled. They often propagate their output
    # to the root anyway.
    "disable_existing_loggers": False,
    "formatters": {"rich": {"datefmt": "[%X]"}},
    "filters": {
        "filter_favicon_requests": {
            "()": "django.utils.log.CallbackFilter",
            "callback": _filter_favicon_requests,
        },
        "filter_static_requests": {
            "()": "django.utils.log.CallbackFilter",
            "callback": _filter_static_requests,
        },
    },
    "handlers": {
        "console": {
            "class": "rich.logging.RichHandler",
            "formatter": "rich",
            "filters": ["filter_favicon_requests", "filter_static_requests"],
        },
    },
    # Existing loggers actually contain direct (non-string) references to existing handlers,
    # so after redefining handlers, all existing loggers must be redefined too
    "loggers": {
        # Configure the root logger to output to the console
        "": {"level": "INFO", "handlers": ["console"], "propagate": False},
        # Django defines special configurations for the "django" and "django.server" loggers,
        # but we will manage all content at the root logger instead, so reset those
        # configurations.
        "django": {
            "handlers": [],
            "level": "NOTSET",
            "propagate": True,
        },
        "django.server": {
            "handlers": [],
            "level": "NOTSET",
            "propagate": True,
        },
    },
}

__all__ = [
    "LOGGING",
]



"""Configure Django's security middleware to use and require HTTPS."""

from datetime import timedelta

SECURE_SSL_REDIRECT = True

SESSION_COOKIE_SECURE = True
CSRF_COOKIE_SECURE = True

# Enable HSTS
SECURE_HSTS_SECONDS = int(timedelta(days=365).total_seconds())
# This is already False by default, but it's important to ensure HSTS is not forced on other
# subdomains which may have different HTTPS practices.
SECURE_HSTS_INCLUDE_SUBDOMAINS = False
# This is already False by default, but per https://hstspreload.org/#opt-in, projects should
# opt-in to preload by overriding this setting. Additionally, all subdomains must have HSTS to
# register for preloading.
SECURE_HSTS_PRELOAD = False

__all__ = [
    "SECURE_SSL_REDIRECT",
    "SESSION_COOKIE_SECURE",
    "CSRF_COOKIE_SECURE",
    "SECURE_HSTS_SECONDS",
    "SECURE_HSTS_INCLUDE_SUBDOMAINS",
    "SECURE_HSTS_PRELOAD",
]



"""
Configure MinioMediaStorage.

This requires the `django-minio-storage` package to be installed.
"""

from urllib.parse import ParseResult

from resonant_settings._env import env

minio_url: ParseResult = env.url("DJANGO_MINIO_STORAGE_URL")
MINIO_STORAGE_USE_HTTPS = minio_url.scheme == "https"
MINIO_STORAGE_ENDPOINT = (
    f"{minio_url.hostname}:{minio_url.port}" if minio_url.port else minio_url.hostname
)
MINIO_STORAGE_ACCESS_KEY = minio_url.username
MINIO_STORAGE_SECRET_KEY = minio_url.password
MINIO_STORAGE_MEDIA_BUCKET_NAME = minio_url.path.lstrip("/")

# Setting this allows MinIO to work through network namespace partitions
# (e.g. when running within Docker Compose)
MINIO_STORAGE_MEDIA_URL: str | None = env.str("DJANGO_MINIO_STORAGE_MEDIA_URL", default=None)

MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET = True
# Make the bucket private to the public
MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY = "NONE"
# Issue signed URLs to provide any read access
MINIO_STORAGE_MEDIA_USE_PRESIGNED = True

__all__ = [
    "MINIO_STORAGE_USE_HTTPS",
    "MINIO_STORAGE_ENDPOINT",
    "MINIO_STORAGE_ACCESS_KEY",
    "MINIO_STORAGE_SECRET_KEY",
    "MINIO_STORAGE_MEDIA_BUCKET_NAME",
    "MINIO_STORAGE_MEDIA_URL",
    "MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET",
    "MINIO_STORAGE_AUTO_CREATE_MEDIA_POLICY",
    "MINIO_STORAGE_MEDIA_USE_PRESIGNED",
]



from allauth.account.adapter import DefaultAccountAdapter
from django.contrib.auth.models import AbstractUser
from django.http import HttpRequest


class EmailAsUsernameAccountAdapter(DefaultAccountAdapter):
    """Automatically populate the username as the email address."""

    def populate_username(self, request: HttpRequest, user: AbstractUser) -> None:
        user.username = user.email



"""
Configure Django Extensions.

This requires the `django-extensions` package to be installed.
"""

SHELL_PLUS_PRINT_SQL = True
SHELL_PLUS_PRINT_SQL_TRUNCATE = None
RUNSERVER_PLUS_PRINT_SQL_TRUNCATE = None

__all__ = [
    "SHELL_PLUS_PRINT_SQL",
    "SHELL_PLUS_PRINT_SQL_TRUNCATE",
    "RUNSERVER_PLUS_PRINT_SQL_TRUNCATE",
]
